﻿#include "mdsst.h"

MDSST::MDSST(QObject *parent) : QObject(parent){}

void MDSST::getNcFileNamesList(const QString &filePath, QStringList &tempFileList)
{
    // 文件路径信息是否为空判断
    if(filePath == NULL){
        QString info = "输入路径有误!";
        emit showInfo(info);
        return;
    } else{
        QString info = "检索当前路径下的文件...";
        emit showInfo(info);
    }

    QDir dir(filePath);
    QStringList nameFilters;
    nameFilters << "*.nc" ;
    tempFileList = dir.entryList(nameFilters, QDir::Files|QDir::Readable, QDir::Name);

    QString info = "文件检索已完成,文件数量:";
    info += tempFileList.size();
    emit showInfo(info);
}

void MDSST::fileRead(QString filePath,NC_OBJ &temp_Obj,QString tempMark)
{
    QByteArray  fpArr       = filePath.toLocal8Bit();
    const char* ncFilePath  = fpArr.data();

    GDALAllRegister();
    // 设置支持中文路径
    CPLSetConfigOption("GDAL_FILENAME_IS_UTF8","NO");
    // 以只读方式打开数据集
    GDALDataset* fileDataset = (GDALDataset*) GDALOpen(ncFilePath,GA_ReadOnly);
    if (fileDataset == NULL)
    {
        QString info = "文件路径有误!";
        emit showInfo(info);
        return;
    }else{
        QString info = "读取文件...";
        emit showInfo(info);
    }

    // 获得数据的字符串，可以打印出来看看自己需要的数据在那
    char** sublist = GDALGetMetadata((GDALDatasetH) fileDataset,"SUBDATASETS");

    int iCount = CSLCount(sublist);
    if(iCount <= 0){
        QString info = filePath + "数据为空!";
        emit showInfo(info);
        GDALClose((GDALDriverH)fileDataset);
        return;
    }

    // 存储数据集信息
    for(int i = 0; sublist[i] != NULL;i++){

        // 打印数据集表述信息
        // qDebug() << sublist[i] << endl;

        if(i%2 != 0){
            continue;
        }

        QString tmpstr = sublist[i];
        tmpstr = tmpstr.mid(tmpstr.indexOf("=")+1);
        QByteArray  tmpstrArr = tmpstr.toLocal8Bit();
        const char* tmpc_str  = tmpstrArr.data();

        QString tmpdsc = sublist[i+1];
        tmpdsc = tmpdsc.mid(tmpdsc.indexOf("=")+1);

        GDALDataset* hTmpDt = (GDALDataset*)GDALOpen(tmpc_str,GA_ReadOnly);//打开该数据

        if (hTmpDt != NULL)
        {
            temp_Obj.dataSet.push_back(tmpc_str);
        }
        if(&temp_Obj.dSDesc != NULL){
            temp_Obj.dSDesc.push_back(tmpdsc);
        }
        GDALClose(hTmpDt);
    }

    temp_Obj.dSAttrInfo.resize(temp_Obj.dataSet.size());
    temp_Obj.dSValue.resize(temp_Obj.dataSet.size());

    QString qtmpdsc = temp_Obj.dSDesc[0];
    QStringList qtmpdsclist = qtmpdsc.split(" ");
    QString dataset_name = qtmpdsclist[1];


    for(int dsIdx = 0; dsIdx < temp_Obj.dataSet.size(); dsIdx++){
        string tempInfo = temp_Obj.dataSet[dsIdx].toStdString();
        GDALDataset* tempDt   = (GDALDataset*)GDALOpen(tempInfo.data(), GA_ReadOnly);

        double adfGeoTransform[6];
        if( tempDt->GetGeoTransform(adfGeoTransform) != CE_None ){
            QString info = filePath + "投影信息不存在!";
            emit showInfo(info);
            return;
        }
        temp_Obj.dSAttrInfo[dsIdx].bCount = tempDt->GetRasterCount();
        temp_Obj.dSValue[dsIdx].resize(tempDt->GetRasterCount());

        /***************************************************************
         *          大隅-吐噶啦海峡
         *          范围：
         *              126 E ~ 135E
         *              26  N ~ 34 N
         * *************************************************************/

        int  iStartX = floor((126 - adfGeoTransform[0]) / adfGeoTransform[1]);
        int  iStartY = floor((34  - adfGeoTransform[3]) / adfGeoTransform[5]);
        int  iSizeX  = floor((135 - 126) / adfGeoTransform[1]);
        int  iSizeY  = floor((26  - 34 ) / adfGeoTransform[5]);

        temp_Obj.dSAttrInfo[dsIdx].orgLocat.x   = adfGeoTransform[0] + iStartX*adfGeoTransform[1] + iStartY*adfGeoTransform[2];
        temp_Obj.dSAttrInfo[dsIdx].orgLocat.y   = adfGeoTransform[3] + iStartX*adfGeoTransform[4] + iStartY*adfGeoTransform[5];
        temp_Obj.dSAttrInfo[dsIdx].pixelSize.x  = adfGeoTransform[1];
        temp_Obj.dSAttrInfo[dsIdx].pixelSize.y  = adfGeoTransform[5];
        temp_Obj.dSAttrInfo[dsIdx].rXSize       = iSizeX;
        temp_Obj.dSAttrInfo[dsIdx].rYSize       = iSizeY;

        QString dataSetNam = temp_Obj.dataSet[dsIdx];
        QString info = "正在解析" + dataSetNam + "...";
        emit showInfo(info);

        int pbnMax = tempDt->GetRasterCount();
        int *pBandMap = new int[pbnMax];
        for(int pbn = 0; pbn<pbnMax;pbn++){
            pBandMap[pbn] = pbn + 1;
        }
        /**********获取数据类型，定义缓存空间大小*******************

        GDALRasterBand* pBand = tempDt->GetRasterBand(1);
        GDALDataType dataType = pBand->GetRasterDataType();

        ********************************************************/

        float* lineData = new float[pbnMax*iSizeX*iSizeY];
        tempDt->RasterIO(GF_Read, iStartX, iStartY, iSizeX, iSizeY,lineData,iSizeX, iSizeY, GDT_Float32, pbnMax, pBandMap, 0, 0, 0);

        int BandNumMax  = pbnMax;
        int iLineMax    = iSizeY;
        int iPixelMax   = iSizeX;

        for(int BandNum = 0;BandNum<BandNumMax;BandNum++){
            int tempBN = BandNum + 1;

            temp_Obj.dSValue[dsIdx][BandNum].resize(iLineMax);
            for (int iLine = 0; iLine < iLineMax; iLine++)
            {
                temp_Obj.dSValue[dsIdx][BandNum][iLine].resize(iPixelMax);
                for (int iPixel = 0; iPixel < iPixelMax; iPixel++)
                {
                    int index = BandNum * iLineMax * iPixelMax + iLine * iPixelMax + iPixel;
                    temp_Obj.dSValue[dsIdx][BandNum][iLine][iPixel] = lineData[index];
                }
            }

            info = "波段" + QString::number(tempBN) + "已完成.";
            emit showInfo(info);
        }

        delete [] lineData;
        lineData = NULL;
        GDALClose((GDALDatasetH)tempDt);

        info = dataSetNam + "解析已完成.";
        emit showInfo(info);
    }

    QString info = filePath + "读取已完成.";
    emit showInfo(info);
    GDALClose(fileDataset);
}

void MDSST::inverDistToPowerItpl(GRD_OBJ itplGRDOBJ,GRD_OBJ &itpledGRDOBJ,double pixResoultion)
{
    // 相关参数信息是否为空判断
    if(pixResoultion == NULL){
        qDebug() << __FUNCTION__
                 << "tempPoint3d: data is empty!"
                 << __LINE__
                 << endl;
        return;
    }else if(itplGRDOBJ.dSAttrInfo.rXSize == NULL){
        qDebug() << __FUNCTION__
                 << "tempGRDOBJ: data is empty!"
                 << __LINE__
                 << endl;
        return;
    }else{
        QString info = "对当前文件进行插值处理...";
        emit showInfo(info);
        qDebug() << "start inverDistToPowerItpl ..." << endl;
    }

    // 有效数据筛选，剔除标记值
    QList<Point3D> filterPoint3d;
    for(int i = 0;i<itplGRDOBJ.dSValue.size();i++){
        for(int j = 0;j<itplGRDOBJ.dSValue[i].size();j++){

            double tempValue = itplGRDOBJ.dSValue[i][j];
            if(tempValue == -999){
                continue;
            }
            Point3D tempP3d;
            tempP3d.y = i + 1;
            tempP3d.x = j + 1;
            tempP3d.z = tempValue;
            filterPoint3d.append(tempP3d);
        }
    }

    // 定义散点总数
    int nPoints = filterPoint3d.size();

    // 定义插值空间范围
    double dfXMin = 0;
    double dfXMax = itplGRDOBJ.dSAttrInfo.rXSize;
    double dfYMin = 0;
    double dfYMax = itplGRDOBJ.dSAttrInfo.rYSize;

    // 定义散点坐标信息
    double* padfX = new double[nPoints];
    double* padfY = new double[nPoints];
    double* padfZ = new double[nPoints];

    for(int i = 0;i<nPoints;i++){
        padfX[i] = filterPoint3d[i].x;
        padfY[i] = filterPoint3d[i].y;
        padfZ[i] = filterPoint3d[i].z;
    }

    int nXSize = (dfXMax-dfXMin)*abs(itplGRDOBJ.dSAttrInfo.pixelSize.x)/pixResoultion;
    int nYSize = (dfYMax-dfYMin)*abs(itplGRDOBJ.dSAttrInfo.pixelSize.y)/pixResoultion;

    // 构造插值算法参数并设置差值参数值
    GDALGridInverseDistanceToAPowerOptions* poOptions = new GDALGridInverseDistanceToAPowerOptions();
    // 权重指数
    poOptions->dfPower   = 2;
    // 搜索椭圆第一半径
    poOptions->dfRadius1 = 20;
    // 搜索椭圆第二半径
    poOptions->dfRadius2 = 15;

    //定义插值后输出数据组
    float* lineData = new float[nXSize*nYSize];

    //反距离权重插值(IDW)
    GDALGridCreate(GGA_InverseDistanceToAPower,poOptions,nPoints,padfX,padfY,padfZ,dfXMin,dfXMax,dfYMin,dfYMax,nXSize,nYSize,GDT_Float32,lineData,NULL,NULL);

    //数据信息整理
    itpledGRDOBJ.dataSet = itplGRDOBJ.dataSet;
    itpledGRDOBJ.dSDesc  = itplGRDOBJ.dSDesc;

    itpledGRDOBJ.dSAttrInfo.bCount = itplGRDOBJ.dSAttrInfo.bCount;
    itpledGRDOBJ.dSAttrInfo.orgLocat = itplGRDOBJ.dSAttrInfo.orgLocat;

    if(itplGRDOBJ.dSAttrInfo.pixelSize.x == abs(itplGRDOBJ.dSAttrInfo.pixelSize.x)){
        itpledGRDOBJ.dSAttrInfo.pixelSize.x = pixResoultion;
    }else{
        itpledGRDOBJ.dSAttrInfo.pixelSize.x = -1 * pixResoultion;
    }
    if(itplGRDOBJ.dSAttrInfo.pixelSize.x == abs(itplGRDOBJ.dSAttrInfo.pixelSize.x)){
        itpledGRDOBJ.dSAttrInfo.pixelSize.y = pixResoultion;
    }else{
        itpledGRDOBJ.dSAttrInfo.pixelSize.y = -1 * pixResoultion;
    }

    itpledGRDOBJ.dSAttrInfo.rXSize = nXSize;
    itpledGRDOBJ.dSAttrInfo.rYSize = nYSize;

    // 插值数据存储
    QVector<QVector<float>> tempValue;
    tempValue.resize(nYSize);
    for(int lineNum = 0;lineNum<nYSize;lineNum++){
        tempValue[lineNum].resize(nXSize);
        for(int pixalNum = 0;pixalNum<nXSize;pixalNum++){
            tempValue[lineNum][pixalNum] = lineData[lineNum*nXSize+pixalNum];
        }
    }

    itpledGRDOBJ.dSValue.append(tempValue);

    delete poOptions;
    delete [] lineData;

    QString info = "插值处理完成.";
    emit showInfo(info);
}

void MDSST::avageOfDailySST(NC_OBJ  temp_Obj,GRD_OBJ &temp_ATInfo,QVector<QVector<float>> &tempAvageTemper)
{
    // 数据存储信息是否为空判断
    if(temp_Obj.dSValue[0].size() == 0){
        return;
    }else if(temp_Obj.dSValue[0][0].size() == 0){
        return;
    }else if(temp_Obj.dSValue[0][0][0].size() == 0){
        return;
    }else{
        //数据值存在
        qDebug() << "check data info completed,no error." << endl;
        QString info = "对插值后的数据进行检查...";
        emit showInfo(info);
    }

    //存储当前的数据信息
    if(temp_ATInfo.dataSet.isEmpty()){
        temp_ATInfo.dataSet = temp_Obj.dataSet[0];
        temp_ATInfo.dSDesc  = temp_Obj.dSDesc[0];

        temp_ATInfo.dSAttrInfo = temp_Obj.dSAttrInfo[0];
        temp_ATInfo.dSAttrInfo.bCount = 1;
    }

    // 插值处理--反距离定权插值
    inverDistToPowerItplProcess(temp_Obj);

    // 日均值计算 处理对象为SST表温数据
    qDebug() << "start calc average temperature of sea..." << endl;
    QString info = "对当前数据进行日平均计算...";
    emit showInfo(info);
    tempAvageTemper.resize(temp_Obj.dSValue[0][0].size());
    for(int i = 0;i < temp_Obj.dSValue[0][0].size();i++){

        tempAvageTemper[i].resize(temp_Obj.dSValue[0][0][i].size());
        for(int j = 0;j < temp_Obj.dSValue[0][0][i].size();j++){

            float   tempNum     = 0;
            int     avgNum      = 0;
            float   tempMark    = 0;

            for(int k = 0;k < temp_Obj.dSValue[0].size();k++){
                float tempValue = temp_Obj.dSValue[0][k][i][j];
                if(tempValue == -999){
                    tempMark = temp_Obj.dSValue[0][k][i][j];
                    continue;
                }
                avgNum  += 1;
                tempNum += temp_Obj.dSValue[0][k][i][j];
            }

            if(avgNum == 0){
                tempAvageTemper[i][j] = tempMark;
            }else{
                tempNum = tempNum/avgNum;
                tempAvageTemper[i][j] = tempNum;
            }

        }
    }
    qDebug() << "Daily mean temperature of sea_surface is received." << endl;
    info = "计算完成.";
    emit showInfo(info);
}

void MDSST::inverDistToPowerItplProcess(NC_OBJ  &temp_Obj)
{
    // 数据存储信息是否为空判断
    if(temp_Obj.dSValue[0].size() == 0){
        qDebug() << __FUNCTION__
                 << "temp_Obj: data is empty!"
                 << __LINE__
                 << endl;
        return;
    }else if(temp_Obj.dSValue[0][0].size() == 0){
        qDebug() << __FUNCTION__
                 << "temp_Obj: data is empty!"
                 << __LINE__
                 << endl;
        return;
    }else if(temp_Obj.dSValue[0][0][0].size() == 0){
        qDebug() << __FUNCTION__
                 << "temp_Obj: data is empty!"
                 << __LINE__
                 << endl;
        return;
    }else{
        //数据值存在
        qDebug() << "check data info completed,ItplProcess start ..." << endl;
    }

    GRD_OBJ   itplGRDOBJ;

    for(int dsIdx = 0; dsIdx < temp_Obj.dataSet.size(); dsIdx++){

        itplGRDOBJ.dataSet = temp_Obj.dataSet[dsIdx];
        itplGRDOBJ.dSDesc  = temp_Obj.dSDesc[dsIdx];

        itplGRDOBJ.dSAttrInfo = temp_Obj.dSAttrInfo[dsIdx];
        itplGRDOBJ.dSAttrInfo.bCount = 1;

        for(int BandNum = 0;BandNum<temp_Obj.dSAttrInfo[dsIdx].bCount;BandNum++){

            itplGRDOBJ.dSValue.append(temp_Obj.dSValue[dsIdx][BandNum]);
            GRD_OBJ itpledGRDOBJ;
            inverDistToPowerItpl(itplGRDOBJ,itpledGRDOBJ,0.25);
            temp_Obj.dSValue[dsIdx][BandNum].clear();
            temp_Obj.dSValue[dsIdx][BandNum].append(itpledGRDOBJ.dSValue);
        }
    }

    qDebug() << "ItplProcess completed." << endl;
}

void MDSST::dataArrangement(GRD_OBJ temp_obj, QVector<QVector<float>> tempAT, QList<Point3D> &tempPL)
{
    // 路径信息是否为空判断
    if(temp_obj.dSAttrInfo.rXSize == NULL){
        qDebug() << __FUNCTION__
                 << "temp_obj:data is empty!"
                 << __LINE__
                 << endl;
        return;
    }else if(tempAT.size() == 0){
        qDebug() << __FUNCTION__
                 << "tempAT:data is empty!"
                 << __LINE__
                 << endl;
        return;
    }else{
        qDebug() << "start dataArrangement ...." << endl;
        QString info = "整理当前成果数据...";
        emit showInfo(info);
    }

    double iXStart = temp_obj.dSAttrInfo.orgLocat.x;
    double iYStart = temp_obj.dSAttrInfo.orgLocat.y;
    double iXPixel = temp_obj.dSAttrInfo.pixelSize.x;
    double iYPixel = temp_obj.dSAttrInfo.pixelSize.y;

    double iXSize  = temp_obj.dSAttrInfo.rXSize;
    double iYSize  = temp_obj.dSAttrInfo.rYSize;

    for(int i = 0;i<iYSize;i++){
        double B = iYStart + i * iYPixel;
        for(int j = 0;j<iXSize;j++){
            double L = iXStart + j * iXPixel;

            Point3D tempP;
            tempP.x = B;
            tempP.y = L;
            tempP.z = tempAT[i][j];

            tempPL.append(tempP);
        }
    }

    QString info = "整理完毕.";
    emit showInfo(info);
    qDebug() << "Point infomation arrange completed!" << endl;
}

void MDSST::createFilePath(QString savePath, QString fileNam, QString &filePath)
{
    QDateTime local(QDateTime::currentDateTime());
    QString localTime = local.toString("yyyyMMdd");
    //文件名称格式为SEAFLUX-OSB-CDR_V02R00_SST_D20060101_C20160824.nc
    QString tempStr = fileNam.split("_")[3];
    filePath = savePath + "/SEAFLUX-OSB-CDR_V02R00_SST_" + tempStr+ "_C" + localTime + ".txt";
}

void MDSST::crateTxtFile(QString fileSavePath, QList<Point3D> tempPointList)
{
    QFile file(fileSavePath);
    /******************************************************************
    if(!file.open(QIODevice::WriteOnly))
    {
        qDebug() << "Open failed, please check the file path!" << endl;
        return;
    }
    ******************************************************************/

    QString info = "写入文件" + fileSavePath;
    emit showInfo(info);
    if(file.open(QFile::ReadWrite)){
        for(int i = 0; i<tempPointList.size();i++){

            /******************************************************************
            QString tempPointInfo = QString::number(tempPointList[i].x) + "," +
                                    QString::number(tempPointList[i].y) + "," +
                                    QString::number(tempPointList[i].z) + "\n";
            file.write(tempPointInfo.toUtf8());
            ******************************************************************/
            QTextStream outPut(&file);
            outPut << tempPointList[i].x
                   <<","
                   << tempPointList[i].y
                   <<","
                   << tempPointList[i].z
                   <<endl;
        }
    }else{
        qDebug() << "Open failed, please check the file path!" << endl;
        return;
    }

    file.close();

    QString fileName = fileSavePath.split("/")[fileSavePath.split("/").size()-1];
    qDebug() << "File "
             << fileName
             << "write completed!"
             << endl;

    info = fileName + "写入完成.";
    emit showInfo(info);
}

void MDSST::startprocess(QString inputFilePath,QString outputFilePath)
{

    inputFilePath  = inputFilePath.replace(QString("\\"),QString("/"));
    outputFilePath = outputFilePath.replace(QString("\\"),QString("/"));

    QStringList fileList;
    getNcFileNamesList(inputFilePath,fileList);

    emit showFileNumInfo(QString::number(fileList.size()));
    for(int i = 0;i<fileList.size();i++){
        QString filePath = inputFilePath + "/" + fileList[i];

        NC_OBJ     Obj;
        GRD_OBJ aTInfo;
        QList<Point3D> pList;
        QVector<QVector<float>> avageTemper;

        QString tempinfo = QString::number((i+0.3)*100/fileList.size()) + "%";
        emit showCompedInfo(tempinfo);
        fileRead(filePath,Obj,"SST");

        tempinfo = QString::number((i+0.6)*100/fileList.size()) + "%";
        emit showCompedInfo(tempinfo);
        avageOfDailySST(Obj,aTInfo,avageTemper);

        tempinfo = QString::number((i+0.9)*100/fileList.size()) + "%";
        emit showCompedInfo(tempinfo);
        dataArrangement(aTInfo,avageTemper,pList);

        QString savePath = "";
        createFilePath(outputFilePath,fileList[i],savePath);
        crateTxtFile(savePath,pList);
        tempinfo = QString::number((i+1)*100/fileList.size()) + "%";
        emit showCompedInfo(tempinfo);

        tempinfo = QString::number(i + 1);
        emit showCompedFileNum(tempinfo);
    }
    emit showInfo("所有文件已处理完毕!");
}
